
// ****************************************************************************
//                                 
//                                  Timer1
//
// ****************************************************************************

#include "include.h"

// current setup of Timer1
u8 Timer1Gen;		// timer 1 generator mode
u8 Timer1Mode;		// timer 1 mode            
u8 Timer1Inx;		// index of prescaler
u16 Timer1SelNow;	// current selected frequency
u16 Timer1Top;		// top value
u16 Timer1Pre;		// prescaler
u32 Timer1Div;		// total divider

// PWM
volatile u32 Timer1Pwm;	// PWM counter
u32 Timer1Add;		// PWM increment

// noise
volatile u32 Timer1Seed; // random seed (can be u16 if USERND32 flag will be used)
volatile u8 Timer1His1; // brown noise history 1
volatile u8 Timer1His2; // brown noise history 2
volatile u8 Timer1His3; // brown noise history 3
volatile u8 Timer1His4; // brown noise history 4
volatile u8 Timer1His5; // brown noise history 5
volatile u8 Timer1His6; // brown noise history 6

// number of frequencies
const u16 Timer1Num[GMODE_NUM] PROGMEM = {
	TIMER1DECNUM,	// GMODE_DEC // A: decimal square
	TIMER1INTNUM,	// GMODE_INT // B: integer square
	TIMER1NOTENUM,	// GMODE_NOTE // C: note square
	TONENUM,	// GMODE_SINE // D: note sine
	TONENUM,	// GMODE_TRI // E: note triangle
	TONENUM,	// GMODE_SAW // F: note sawtooth
	3,		// GMODE_NOISE // G: noise
};

// selected frequencies
u16 Timer1Sel[GMODE_NUM] = {
	9,	// GMODE_DEC // A: decimal square
	28,	// GMODE_INT // B: integer square
	120,	// GMODE_NOTE // C: note square
	120,	// GMODE_SINE // D: note sine
	120,	// GMODE_TRI // E: note triangle
	120,	// GMODE_SAW // F: note sawtooth
	1,	// GMODE_NOISE // G: noise
};

// prescaler values
const u16 Timer1PreVal[8] PROGMEM = { 0, 1, 8, 64, 256, 1024, 0, 0 };

// set Timer1 square mode
void Timer1Square(u16 top, u8 inx)
{
	// Timer 1 mode
	Timer1Mode = ((inx & FASTFLAG) != 0) ? TMODE_FAST : TMODE_SLOW;

	// prescaler
	inx &= PRESMASK;
	Timer1Inx = inx;
	Timer1Pre = pgm_read_word(&Timer1PreVal[inx]);

	// top value
	Timer1Top = top;

	// total divider
	u32 d = Timer1Pre * ((u32)top+1);

	// fast mode
	if (Timer1Mode == TMODE_FAST)
	{
		// set middle
		OCR1A = (top+1)/2-1;

		// set TOP
		ICR1 = top;

		// set fast PWM mode 14: set OC1A at OCR1A compare match, clear OC1A at TOP=ICR1, set prescaler
		TCCR1A = BIT(WGM11) | BIT(COM1A1) | BIT(COM1A0);
		TCCR1B = BIT(WGM12) | BIT(WGM13) | inx;
	}

	// slow mode
	else
	{
		// set TOP
		OCR1A = top;

		// set fast PWM mode 15: toggle OC1A at TOP=OCR1A compare match, set prescaler
		TCCR1A = BIT(WGM11) | BIT(WGM10) | BIT(COM1A0);
		TCCR1B = BIT(WGM12) | BIT(WGM13) | inx;

		d *= 2;
	}

	Timer1Div = d;
}

// set Timer1 PWM mode
void Timer1SetPwm(s16 sel)
{
	// PW mode
	Timer1Mode = TMODE_PWM;

	// PWM phase counter
	Timer1Pwm = 0;

	// PWM phase increment
	Timer1Add = pgm_read_dword(&ToneSineInc[sel]);

	// set first TOP
	OCR1A = 128;

	// set fast PWM mode 8-bit (mode 5), no prescaler, set HIGH at TOP, set LOW at compare match
	//  - Noise mode requires slower interrupt: phase correct PWM 8-bit (mode 1)
	if (Timer1Gen == GMODE_NOISE)
	{
		TCCR1A = BIT(COM1A1) | BIT(WGM10);
		TCCR1B = 1;
	}
	else
	{
		TCCR1A = BIT(COM1A1) | BIT(WGM10);
		TCCR1B = BIT(WGM12) | 1;
	}

	// enable interrupt on TOP
	TIMSK |= BIT(TOIE1);
}

// setup Timer1 (gen=generator mode, sel=frequency)
void Timer1Set(s8 gen, s16 sel)
{
	// set port to output mode
	T1DDR |= T1BIT;

	// disable interrupts
	cli();

	// stop Timer1
	TCCR1B = 0;
	TCCR1A = 0;
	TCNT1 = 0;

	// disable interrupt
	TIMSK &= ~BIT(TOIE1);

	// reset interrupt flag
	TIFR = BIT(TOV1);

	// save new setup
	Timer1Gen = gen;
	Timer1Sel[gen] = sel;
	Timer1SelNow = sel;

	// switch generator
	switch (gen)
	{
	case GMODE_DEC: // A: decimal square
		Timer1Square(pgm_read_word(&Timer1DecTop[sel]),
			pgm_read_byte(&Timer1DecInx[sel]));
		break;

	case GMODE_INT: // B: integer square
		Timer1Square(pgm_read_word(&Timer1IntTop[sel]),
			eeprom_read_byte(&Timer1IntInx[sel]));
		break;

	case GMODE_NOTE: // C: note square
		Timer1Square(pgm_read_word(&Timer1NoteTop[sel]),
			pgm_read_byte(&Timer1NoteInx[sel]));
		break;

	case GMODE_SINE: // D: note sine
	case GMODE_TRI: // E: note triangle
	case GMODE_SAW: // F: note sawtooth
	case GMODE_NOISE: // G: noise
		Timer1SetPwm(sel);
		break;
	}

	// enable interrupts
	sei();
}

// set Timer1 generator
void Timer1SetGen(s8 gen)
{
	// overflow generator
	if (gen >= (s8)GMODE_NUM) gen -= GMODE_NUM;
	if (gen < 0) gen += GMODE_NUM;

	// setup Timer1
	Timer1Set(gen, Timer1Sel[gen]);
}

// set Timer1 frequency
void Timer1SetSel(s16 sel)
{
	// prepare number of frequencies
	s8 gen = Timer1Gen;
	u16 num = pgm_read_word(&Timer1Num[gen]);

	// overflow frequency
	if (sel >= (s16)num) sel -= num;
	if (sel < 0) sel += num;

	// setup Timer1
	Timer1Set(gen, sel);
}
