
// ****************************************************************************
//                                 
//                                  Timer2
//
// ****************************************************************************

#include "include.h"

// current setup of Timer2
u8 Timer2Gen;		// timer 2 generator mode
u8 Timer2Mode;		// timer 2 mode            
u8 Timer2Inx;		// index of prescaler
u8 Timer2Top;		// top value
u16 Timer2SelNow;	// current selected frequency
u16 Timer2Pre;		// prescaler
u8 Timer2Int;		// interrupt divider
u32 Timer2Div;		// total divider

volatile u8 Timer2Cnt;	// interrupt counter

// PWM
volatile u32 Timer2Pwm;	// PWM counter
u32 Timer2Add;		// PWM increment

// noise
volatile u32 Timer2Seed; // random seed (can be u16 if USERND32 flag will be used)
volatile u8 Timer2His1; // brown noise history 1
volatile u8 Timer2His2; // brown noise history 2
volatile u8 Timer2His3; // brown noise history 3
volatile u8 Timer2His4; // brown noise history 4
volatile u8 Timer2His5; // brown noise history 5
volatile u8 Timer2His6; // brown noise history 6

// number of frequencies
const u16 Timer2Num[GMODE_NUM] PROGMEM = {
	TIMER2DECNUM,	// GMODE_DEC // A: decimal square
	TIMER2INTNUM,	// GMODE_INT // B: integer square
	TIMER2NOTENUM,	// GMODE_NOTE // C: note square
	TONENUM,	// GMODE_SINE // D: note sine
	TONENUM,	// GMODE_TRI // E: note triangle
	TONENUM,	// GMODE_SAW // F: note sawtooth
	3,		// GMODE_NOISE // G: noise
};

// selected frequencies
u16 Timer2Sel[GMODE_NUM] = {
	9,	// GMODE_DEC // A: decimal square
	27,	// GMODE_INT // B: integer square
	120,	// GMODE_NOTE // C: note square
	120,	// GMODE_SINE // D: note sine
	120,	// GMODE_TRI // E: note triangle
	120,	// GMODE_SAW // F: note sawtooth
	1,	// GMODE_NOISE // G: noise
};

// prescaler values
const u16 Timer2PreVal[8] PROGMEM = { 0, 1, 8, 32, 64, 128, 256, 1024 };

// Timer2 compare
ISR(TIMER2_COMP_vect)
{
	u8 t = Timer2Cnt;
	t--;
	if (t == 0)
	{
		T2PORT ^= T2BIT;
		t = Timer2Int;
	}
	Timer2Cnt = t;
}

// set Timer2 square mode
void Timer2Square(u8 top, u8 inx, u8 inter)
{
	// Timer 2 mode
	Timer2Mode = (inter > 1) ? TMODE_CNT : TMODE_SLOW;

	// prescaler
	Timer2Inx = inx;
	Timer2Pre = pgm_read_word(&Timer2PreVal[inx]);

	// top value
	Timer2Top = top;

	// interrupt
	Timer2Int = inter;
	Timer2Cnt = inter;

	// total divider
	u32 d = Timer2Pre * ((u32)top+1) * inter * 2;

	// set TOP
	OCR2 = top;

	// slow mode
	if (Timer2Mode == TMODE_SLOW)
	{
		// set CTC mode: toggle OC2 at TOP=OCR2 compare match, set prescaler
		TCCR2 = BIT(COM20) | BIT(WGM21) | inx;
	}

	// counter mode
	else
	{
		// set CTC mode: toggle OC2 at interrupts, TOP=OCR2, set prescaler
		TCCR2 = BIT(WGM21) | inx;

		// enable interrupt
		TIMSK |= BIT(OCIE2);
	}

	Timer2Div = d;
}

// set Timer2 PWM mode
void Timer2SetPwm(s16 sel)
{
	// PW mode
	Timer2Mode = TMODE_PWM;

	// PWM phase counter
	Timer2Pwm = 0;

	// PWM phase increment
	Timer2Add = pgm_read_dword(&ToneSineInc[sel]);

	// set first TOP
	OCR2 = 128;

	// set fast PWM mode 8-bit (mode 3), no prescaler, set HIGH at TOP, set LOW at compare match
	//  - Noise mode requires slower interrupt: phase correct PWM 8-bit (mode 1)
	if (Timer2Gen == GMODE_NOISE)
	{
		TCCR2 = BIT(COM21) | BIT(WGM20) | 1;
	}
	else
	{
		TCCR2 = BIT(COM21) | BIT(WGM21) | BIT(WGM20) | 1;
	}

	// enable interrupt on TOP
	TIMSK |= BIT(TOIE2);
}

// setup Timer2 (gen=generator mode, sel=frequency)
void Timer2Set(s8 gen, s16 sel)
{
	// set port to output mode
	T2DDR |= T2BIT;

	// disable interrupts
	cli();

	// stop Timer2
	TCCR2 = 0;
	TCNT2 = 0;

	// disable interrupt
	TIMSK &= ~(BIT(OCIE2) | BIT(TOIE2));

	// reset interrupt flag
	TIFR = BIT(OCF2) | BIT(TOV2);

	// save new setup
	Timer2Gen = gen;
	Timer2Sel[gen] = sel;
	Timer2SelNow = sel;

	// switch generator
	switch (gen)
	{
	case GMODE_DEC: // A: decimal square
		Timer2Square(pgm_read_byte(&Timer2DecTop[sel]),
			pgm_read_byte(&Timer2DecInx[sel]),
			pgm_read_byte(&Timer2DecInt[sel]));
		break;

	case GMODE_INT: // B: integer square
		Timer2Square(eeprom_read_byte(&Timer2IntTop[sel]),
			eeprom_read_byte(&Timer2IntInx[sel]),
			eeprom_read_byte(&Timer2IntInt[sel]));
		break;

	case GMODE_NOTE: // C: note square
		Timer2Square(pgm_read_byte(&Timer2NoteTop[sel]),
			pgm_read_byte(&Timer2NoteInx[sel]),
			pgm_read_byte(&Timer2NoteInt[sel]));
		break;

	case GMODE_SINE: // D: note sine
	case GMODE_TRI: // E: note triangle
	case GMODE_SAW: // F: note sawtooth
	case GMODE_NOISE: // G: noise
		Timer2SetPwm(sel);
		break;
	}

	// enable interrupts
	sei();
}

// set Timer2 generator
void Timer2SetGen(s8 gen)
{
	// overflow generator
	if (gen >= (s8)GMODE_NUM) gen -= GMODE_NUM;
	if (gen < 0) gen += GMODE_NUM;

	// setup Timer2
	Timer2Set(gen, Timer2Sel[gen]);
}

// set Timer2 frequency
void Timer2SetSel(s16 sel)
{
	// prepare number of frequencies
	s8 gen = Timer2Gen;
	u16 num = pgm_read_word(&Timer2Num[gen]);

	// overflow frequency
	if (sel >= (s16)num) sel -= num;
	if (sel < 0) sel += num;

	// setup Timer2
	Timer2Set(gen, sel);
}
