
// ****************************************************************************
//                                 
//                                    Display
//
// ****************************************************************************

#include "include.h"

// prefix table (exp: -12, -9, -6, -3, 0, +3, +6, +9)
const u8 PrefixTab[NUM_PREFIXES] PROGMEM = {'p', 'n', CHAR_MICRO, 'm', 0, 'k', 'M', 'G'};

// current cursor position (can point outside display area)
s8 DispPos = 0; // current display position
s8 DispRow = 0; // current display row

// temporary output buffer
u8 OutBuf[OUTBUF_SIZE];

// set display output position (can draw out of display) (takes 95 us)
void DispSetPos(s8 row, s8 pos)
{
	// check if position is valid and if it has been changed
	if ((pos >= 0) && (pos < LCD_COLNUM) &&
		(row >= 0) && (row < LCD_ROWNUM) &&
		((pos != DispPos) || (row != DispRow)))
		LCD_DDRAM(row*0x40 + pos); // set new cursor position

	// save new position
	DispPos = pos;
	DispRow = row;
}

// output character to current position (takes 95 us)
void DispChar(u8 ch)
{
	// check if current position is valid
	if ((DispPos >= 0) && (DispPos < LCD_COLNUM) &&
		(DispRow >= 0) && (DispRow < LCD_ROWNUM))
		LCD_WriteData(ch); // output character

	// increase cursor position
	DispPos++;
}

// display ASCIIZ text string from ROM (takes 95 us per character)
void DispTextROM(const u8* text)
{
	u8 ch;
	for(;;)
	{
		ch = pgm_read_byte(text);
		if (ch == 0) break;
		DispChar(ch);
		text++;
	}
}

// display text string from RAM, with length (takes 95 us per character)
void DispTextRAM(const u8* text, u8 len)
{
	for (; len > 0; len--)
	{
		DispChar(*text);
		text++;
	}
}

// display ASCIIZ text string from EEPROM (takes 95 us per character)
/*void DispTextEEPROM(const u8* text)
{
	u8 ch;
	for(;;)
	{
		ch = eeprom_read_byte(text);
		if (ch == 0) break;
		DispChar(ch);
		text++;
	}
}
*/
// display space character (takes 95 us)
void DispSpace()
{
	DispChar(' ');
}

// tabulate on position
void DispTab(u8 pos)
{
	while (DispPos < pos) DispSpace();
}

// clear rest of current line
void DispClearRestLine()
{
	while (DispPos < LCD_COLNUM) DispSpace();
}

// get number of digits of unsigned integer value
u8 NumberOfDigits(u32 val)
{
	u32 ref = 9;
	u8 len = 1;
	while (ref < val)
	{
		len++;
		if (ref == 999999999) break;
		ref = ref*10 + 9;
	}
	return len;
}

// decode unsigned number into end of OutBuf, returns number of digits
u8 DecNum(u32 val)
{
	u8* d = &OutBuf[OUTBUF_SIZE-1];
	u8 len = 0;
	u8 ch;
	u32 k;
	do {
		k = val / 10;
		ch = (u8)(val - k*10);
		*d = ch + '0';
		d--;
		len++;
		val = k;
	} while (val > 0);
	return len;
}

// display unsigned integer (takes approx. 800 us)
void DispUInt(u32 val)
{
	u8 len = DecNum(val);
	DispTextRAM(&OutBuf[OUTBUF_SIZE - len], len);
}

// display signed integer (sign=always display sign character, even positive) (takes approx. 800 us)
void DispInt(s32 val, Bool sign)
{
	if (val < 0)
	{
		DispChar('-');
		val = -val;
	}
	else
		if (sign) DispChar('+');

	DispUInt(val);
}

// display unsigned value vith decimal places, prefix and unit (takes approx. 1 ms)
// val=value
// ex=exponent related to base unit, in range -18..+8)
// unit=unit character, 0=none
void DispUVal(u32 val, s8 ex, u8 unit)
{
	u8 pref = 0; // prefix (0=none)
	u8 off = 0; // exponent offset to next 10^3 step
	u8 inx; // index ID
	u8 len; // string length
	u8* s; // text pointer

	// normalize value to 6 digits
	while (val > VALMAX)
	{
		val += 5; // rounding
		val /= 10; // scale value down by 10
		ex++; // increase exponent
	}

	if (val != 0)
	{
		while (val < VALMIN)
		{
			val *= 10;
			ex--;
		}
	}

	// determine prefix and offset
	ex += 11+DIGITS; // shift exponent to be >=0
	inx = ex / 3; // number of 10^3 steps
	off = ex - inx*3; // offset of lower 10^3 steps
	off = DIGITS-1 - off; // reverse offset 1..3

	// get prefix (inx: 0='p', 1='n', ... 7='G')
	pref = pgm_read_byte(PrefixTab + inx);

	// decode number into buffer
	len = DecNum(val);
	s = &OutBuf[OUTBUF_SIZE - len];

	// dot position
	ex = len - off - 1;

	// prepend zero 0.
	if (ex < 0)
	{
		DispChar('0');
		DispChar('.');
		if (ex < -1) DispChar('0'); // 0.0
		if (ex < -2) DispChar('0'); // 0.00
	}

	// display value
	for (inx = 0; inx < len; inx++)
	{
		DispChar(*s++);
		if (inx == ex) DispChar('.');
	}
	
	// display prefix and unit
//	if ((pref != 0) || (unit != 0)) DispSpace();
	if (pref != 0) DispChar(pref);
	if (unit != 0) DispChar(unit);
}
